from time import sleep
import zmq

class Connections():
    """
    implements ZMQ connections between process for PRODUCER and STAGER and CONSUMER
    """
    def __init__(self, main_connection_name, connections=None):
        """
        Parameters
        ----------
        connections : dict
        main_connection_name : str
            Default next step name. Used to send data when destination is not provided
        """
        self.connections = connections or {}
        self.sockets=dict()
        self.context = zmq.Context()
        self.main_out_socket = None
        self.main_connection_name = main_connection_name


    def close_connections(self):
        """
        Close all zmq socket connections
        """
        for sock in self.sockets.values():
            sock.close()

    def get_destination_msg_from_result(self,result):
        """
        If result is a tuple, check if last tuple elem is a valid next step name.
        If yes, return a destination defined to  the last tuple elem and send
        result without the destination
        If no return None as destination

        Parameters
        ----------
        result : any type
            value to send. If type(result) is tuple, it can contain next step name)
        Returns
        --------
        tulpe conaining two elements: msg and destination

        """
        if isinstance(result,tuple):
            # look is last tuple elem is a valid next step
            if result[-1] in self.connections.keys():
                destination = result[-1]
                if len(result [:-1]) == 1:
                    msg = result [:-1][0]
                else:
                    msg = result[:-1]
                return msg,destination
            else:
                return result,None
        else:
            return result,None

    def send_msg(self,msg,destination_step_name=None):
        """
        Send a message thanks to ZMQ

        Parameters
        ----------
        msg: a Pickle.dump message
        destination_step_name: str
            msg will be send to corresponding step
        """
        send=False
        if not destination_step_name :
            socket  = self.main_out_socket
        else:
            socket = self.sockets[destination_step_name]
        while not send:
            socket.send_pyobj(msg)
            request = socket.recv()
            if request == b'OK':
                send = True
            else:
                sleep(0.1)

    def init_connections(self):
        """
        Initialise zmq sockets.
        Because this class is s Process, This method must be call in the run
         method to be hold by the correct process.
        """
        for name,connection in self.connections.items():
            self.sockets[name] = self.context.socket(zmq.REQ)
            try:
                self.sockets[name].connect('tcp://localhost:' + connection)
                if self.main_connection_name == name:
                    self.main_out_socket = self.sockets[name]
            except zmq.error.ZMQError as e:
                print(' {} : tcp://localhost:{}'
                               .format(e,  connection))
                return False
        return True



import sys
from ctapipe.core import Tool
from traitlets import Integer
from ctapipe.flow.gui.main_window import ModuleApplication

class PipeGui(Tool):
    """
    Main program that lunches GUI
    This GUI communicate with pipeline  elements (stager, consumer, producer,
     router) tahnks to ZMQ library PUB/SUB message
    Every second, pipeline send its full configuration:
     - producer
     - levels of stages
     - consumer
     - router / queue
     Each time a producer, stager , consumer or router changes, it send its status
     to this GUI
    """
    description = "run stages in multiprocess pipeline"
    port = Integer(5565, help='GUI port for pipelien connection').tag(
        config=True, allow_none=True)
    def start(self):
        ModuleApplication(sys.argv, self.port)
    def setup(self):
        pass
    def finish(self):
        pass

def main():
    gui = PipeGui()
    gui.run()

if __name__ == "__main__":
    main()



from ctapipe.analysis.camera.chargeresolution import ChargeResolutionCalculator
import numpy as np


def test_add_charges():
    chargeres = ChargeResolutionCalculator(None, None)
    true_charge = np.arange(100)
    measured_charge = np.arange(100)
    chargeres.add_charges(true_charge, measured_charge)
    variation_hist = chargeres.variation_hist

    assert(variation_hist[0][0] == 1)


# def test_add_source():
    # Cannot currently test - no test file has true charge


def test_get_charge_resolution():
    chargeres = ChargeResolutionCalculator(None, None, binning=None)
    true_charge = np.arange(100)
    measured_charge = np.arange(100)
    chargeres.add_charges(true_charge, measured_charge)
    true_charge, chargeres, chargeres_error, \
        scaled_chargeres, scaled_chargeres_error = \
        chargeres.get_charge_resolution()

    assert(true_charge[0] == 1)
    assert(chargeres[0] == 1)
    assert(round(chargeres_error[0], 7) == 0.7071068)
    assert(round(scaled_chargeres[0], 7) == 0.5550272)
    assert(round(scaled_chargeres_error[0], 7) == 0.3924635)


def test_get_binned_charge_resolution():
    chargeres = ChargeResolutionCalculator(None, None)
    true_charge = np.arange(100)
    measured_charge = np.arange(100)
    chargeres.add_charges(true_charge, measured_charge)
    true_charge, chargeres, chargeres_error, \
        scaled_chargeres, scaled_chargeres_error = \
        chargeres.get_charge_resolution()

    assert(true_charge[0] == 1)
    assert(chargeres[0] == 1)
    assert(round(chargeres_error[0], 7) == 0.7071068)
    assert(round(scaled_chargeres[0], 7) == 0.5550272)
    assert(round(scaled_chargeres_error[0], 7) == 0.3924635)


def test_limit_curves():
    value = ChargeResolutionCalculator.limit_curves(1, 2, 3, 4, 5)
    assert(round(value, 7) == 6.78233)


def test_requirement():
    value = ChargeResolutionCalculator.requirement(1)
    value2 = ChargeResolutionCalculator.requirement(np.arange(1, 10))[-1]
    assert(round(value, 7) == 2.0237963)
    assert(round(value2, 7) == 0.4501817)


def test_goal():
    value = ChargeResolutionCalculator.goal(1)
    value2 = ChargeResolutionCalculator.goal(np.arange(1, 10))[-1]
    assert(round(value, 7) == 1.8017134)
    assert(round(value2, 7) == 0.4066657)



import numpy as np
import astropy.units as u
import scipy.stats

from ctapipe.analysis.sensitivity import SensitivityPointSource
from ctapipe.analysis.sensitivity import (check_min_n, check_background_contamination,
                                          cr_background_rate, e_minus_2, sigma_lima,
                                          make_mock_event_rate, crab_source_rate)


def test_check_min_n(n_1=2, n_2=2):
    n = [n_1, n_2]

    min_n = 10

    scale = check_min_n(n, alpha=1, min_n=min_n)

    assert sum(n) >= min_n


def test_check_background_contamination(n_1=2, n_2=2):
    n = [n_1, n_2]

    max_background_ratio = .1

    scale = check_background_contamination(n, alpha=1,
                                           max_background_ratio=max_background_ratio)

    assert n[1] / sum(n) <= max_background_ratio


def test_SensitivityPointSource_effective_area():

    # areas used in the event generator
    gen_area_g = np.pi * (1000 * u.m)**2
    gen_area_p = np.pi * (2000 * u.m)**2

    # energy list of "selected" events
    energy_sel_gammas = np.logspace(2, 6, 200, False) * u.TeV
    energy_sel_electr = np.logspace(2, 6, 200, False) * u.TeV
    energy_sel_proton = np.logspace(2, 6, 400, False) * u.TeV

    # energy list of "generated" events
    energy_sim_gammas = np.logspace(2, 6, 400, False) * u.TeV
    energy_sim_electr = np.logspace(2, 6, 400, False) * u.TeV
    energy_sim_proton = np.logspace(2, 6, 800, False) * u.TeV

    # binning for the energy histograms
    energy_edges = np.logspace(2, 6, 41) * u.TeV

    # energy histogram for the generated events
    energy_sim_hist_gamma = np.histogram(energy_sim_gammas,
                                         bins=energy_edges)[0]
    energy_sim_hist_elect = np.histogram(energy_sim_electr,
                                         bins=energy_edges)[0]
    energy_sim_hist_proton = np.histogram(energy_sim_proton,
                                          bins=energy_edges)[0]

    # constructer gets fed with energy lists and desired energy binning
    sens = SensitivityPointSource(
        mc_energies={'g': energy_sel_gammas,
                     'p': energy_sel_proton,
                     'e': energy_sel_electr},
        energy_bin_edges={'g': energy_edges,
                          'p': energy_edges,
                          'e': energy_edges})

    # effective areas
    sens.get_effective_areas(
        generator_energy_hists={'g': energy_sim_hist_gamma,
                                'p': energy_sim_hist_proton,
                                'e': energy_sim_hist_elect},
        generator_areas={'g': gen_area_g,
                         'p': gen_area_p,
                         'e': gen_area_g})

    # midway result are the effective areas
    eff_a = sens.effective_areas
    eff_area_g, eff_area_p, eff_area_e = eff_a['g'], eff_a['p'], eff_a['e']

    # "selected" events are half of the "generated" events
    # so effective areas should be half of generator areas, too
    np.testing.assert_allclose(eff_area_g.value, gen_area_g.value / 2)
    np.testing.assert_allclose(eff_area_p.value, gen_area_p.value / 2)


def test_SensitivityPointSource_sensitivity_MC():

    alpha = 1.
    n_sig = 20
    n_bgr = 10
    # one bin with the bin-centre at 1 TeV
    energy_bin_edges = np.array([.1, 10]) * u.TeV
    energies_sig = np.ones(n_sig)
    energies_bgr = np.ones(n_bgr)

    sens = SensitivityPointSource(reco_energies={'s': energies_sig * u.TeV,
                                                 'b': energies_bgr * u.TeV})

    # sens.event_weights = {'s': energies_sig, 'b': energies_bgr}
    sensitivity = sens.get_sensitivity(alpha=alpha, signal_list=("s"), mode="MC",
                                       sensitivity_source_flux=crab_source_rate,
                                       sensitivity_energy_bin_edges=energy_bin_edges,
                                       min_n=0, max_background_ratio=1)

    # the ratio between sensitivity and reference flux (i.e. from the crab nebula) should
    # be the scaling factor that needs to be applied to n_sig to produce a 5 sigma result
    # in the lima formula
    ratio = sensitivity["Sensitivity"][0] / (crab_source_rate(1 * u.TeV)).value
    np.testing.assert_allclose(
        [sigma_lima(ratio * n_sig + alpha * n_bgr, n_bgr, alpha)], [5])


def test_SensitivityPointSource_sensitivity_data():

    alpha = 1.
    n_on = 30
    n_off = 10
    # one bin with the bin-centre at 1 TeV
    energy_bin_edges = np.array([.1, 10]) * u.TeV
    energies_sig = np.ones(n_on)
    energies_bgr = np.ones(n_off)

    sens = SensitivityPointSource(reco_energies={'s': energies_sig * u.TeV,
                                                 'b': energies_bgr * u.TeV})

    sensitivity = sens.get_sensitivity(alpha=alpha, signal_list=("s"), mode="data",
                                       sensitivity_source_flux=crab_source_rate,
                                       sensitivity_energy_bin_edges=energy_bin_edges,
                                       min_n=0, max_background_ratio=1)

    # the ratio between sensitivity and reference flux (i.e. from the crab nebula) should
    # be the scaling factor that needs to be applied to n_sig=n_on-n_off*alpha to produce
    # a 5 sigma result  in the lima formula
    ratio = sensitivity["Sensitivity"][0] / (crab_source_rate(1 * u.TeV)).value
    np.testing.assert_allclose([sigma_lima(ratio * (n_on - n_off * alpha) + n_off * alpha,
                                           n_off, alpha)], [5])


def test_generate_toy_timestamps():
    from scipy import signal
    gamma_light_curve = signal.gaussian(5000, 700) * 1000

    tmin, tmax = 0, 5

    n_throws = 500

    time_stamps = SensitivityPointSource.\
        generate_toy_timestamps(light_curves={'g': gamma_light_curve,
                                              'p': n_throws},
                                time_window=(tmin, tmax))

    assert len(time_stamps['g']) == int(np.sum(gamma_light_curve))
    assert len(time_stamps['p']) == n_throws
    for st in time_stamps.values():
        assert (np.min(st) >= tmin) and (np.max(st) <= tmax)
        np.testing.assert_allclose(np.mean(st), 2.5, atol=.25)


def test_draw_events_with_flux_weight():

    np.random.seed(1)

    e_min = 20
    e_max = 50
    n_draws = 10000

    energy = np.random.uniform(e_min, e_max, 50000)

    gamma_old = 0  # since events are drawn from np.random.uniform
    gamma_new = 3

    # this is a simple approach on how to determine event weights (i.e. probabilities to
    # get picked in the  random draw) to generate a set of events with a different energy
    # spectrum
    #
    # Parameters
    # ----------
    # energy : numpy array
    #     the energies of the MC events
    # gamma_new : float
    #     the spectral index the drawn set of events is supposed to follow
    # gamma_old : float
    #     the spectral index the MC events have been generated with
    #     this is zero here but usually 2 in MC generators
    #
    # Returns
    # -------
    # weights : numpy array
    #     list of event weights normalised to 1
    #     to be used as PDF in `np.random.choice`
    energy_weights = energy**(gamma_old - gamma_new)
    energy_weights /= np.sum(energy_weights)

    indices = SensitivityPointSource.draw_events_from_flux_weight(
        {'g': energy},
        {'g': energy_weights},
        {'g': n_draws})

    assert len(energy[indices['g']]) == n_draws
    # TODO come up with more tests...


def test_draw_events_from_flux_histogram():

    np.random.seed(2)

    e_min = 20
    e_max = 50
    n_bins = 30
    n_draws = 1000

    energy_edges = np.linspace(e_min, e_max, n_bins + 1, True) * u.TeV
    energy = np.random.uniform(e_min, e_max, 50000) * u.TeV

    target_distribution = make_mock_event_rate(lambda e: (e / u.TeV)**-3,
                                               energy_edges,
                                               norm=n_draws, log_e=False)

    indices = SensitivityPointSource.draw_events_from_flux_histogram(
        {'g': energy},
        {'g': target_distribution},
        {'g': energy_edges})

    hist, _ = np.histogram(energy[indices['g']], bins=energy_edges[::])

    # checking the χ² between `target_distribution` and the drawn one
    chisquare = scipy.stats.chisquare(target_distribution, hist)[0]
    # the test that the reduced χ² is close to 1 (tollorance of 1)
    np.testing.assert_allclose([chisquare / n_bins], [1], atol=0.1)



from os.path import dirname, exists, splitext, basename
from os import makedirs
from math import ceil, floor
from matplotlib import pyplot as plt
from math import log10
import warnings
import numpy as np
from matplotlib.colors import LogNorm
from traitlets import Dict, List, Unicode
from ctapipe.core import Tool, Component
from ctapipe.analysis.camera.chargeresolution import ChargeResolutionCalculator


class ChargeResolutionVariationPlotter(Component):
    name = 'ChargeResolutionVariationPlotter'

    output_path = Unicode(None, allow_none=True,
                          help='Output path to save the '
                               'plot.').tag(config=True)

    def __init__(self, config, tool, **kwargs):
        """
        Calculator of charge resolution.

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        reductor : ctapipe.calib.camera.reductors.Reductor
            The reductor to use to reduce the waveforms in the event.
            By default no data volume reduction is applied, and the dl0 samples
            will equal the r1 samples.
        kwargs
        """
        super().__init__(config=config, parent=tool, **kwargs)

        try:
            if self.output_path is None:
                raise ValueError
        except ValueError:
            self.log.exception('Please specify an output path')
            raise

        self.fig = plt.figure(figsize=(20, 8))
        self.ax_l = self.fig.add_subplot(121)
        self.ax_r = self.fig.add_subplot(122)

        self.fig.subplots_adjust(left=0.05, right=0.95, wspace=0.6)

        self.legend_handles = []
        self.legend_labels = []

    def plot_hist(self, hist, xedges, yedges):

        hist[hist == 0.0] = np.nan

        fig = plt.figure(figsize=(10, 7))
        ax = fig.add_subplot(111)
        ax.set_title(splitext(basename(self.output_path))[0])
        x, y = np.meshgrid(xedges, yedges)
        x = np.power(10, x)
        y = np.power(10, y)
        hist_mask = np.ma.masked_where(np.isnan(hist), hist)
        im = ax.pcolormesh(x, y, hist_mask, norm=LogNorm(),
                           cmap=plt.cm.viridis)
        cb = plt.colorbar(im)
        ax.set_aspect('equal')
        ax.grid()
        ax.set_xscale('log')
        ax.set_yscale('log')
        ax.set_xlabel(r'True Charge $Q_T$ (p.e.)')
        ax.set_ylabel(r'Measured Charge $Q_M$ (p.e.)')
        cb.ax.set_ylabel("Count")

        line = np.linspace(*ax.get_xlim(), 100)
        ax.plot(line, line, c='0.75', ls='--')

        # Add minor ticks
        lmin = floor(log10(hist_mask.min()))
        lmax = ceil(log10(hist_mask.max()))
        logticks = np.tile(np.arange(lmin, 10), lmax) * (
            np.power(10, np.arange(lmax * 10) // 10))
        logticks = im.norm(logticks[(logticks != 0) &
                                    (logticks >= hist_mask.min()) &
                                    (logticks <= hist_mask.max())])
        cb.ax.yaxis.set_ticks(logticks, minor=True)
        cb.ax.yaxis.set_tick_params(which='minor', length=5)
        cb.ax.tick_params(length=10)

        output_dir = dirname(self.output_path)
        if not exists(output_dir):
            self.log.info("[output] Creating directory: {}".format(output_dir))
            makedirs(output_dir)
        self.log.info("[output] {}".format(self.output_path))
        warnings.filterwarnings("ignore", module="matplotlib")
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            fig.savefig(self.output_path, bbox_inches='tight')


class ChargeResolutionVariationViewer(Tool):
    name = "ChargeResolutionVariationViewer"
    description = "Plot the charge resolution from " \
                  "ChargeResolutionCalculator objects restored via " \
                  "pickled dictionaries."

    input_path = Unicode(None, allow_none=True,
                         help='Path to the pickle file produced from'
                              'ChargeResolutionCalculator.save()'
                              '').tag(config=True)

    aliases = Dict(dict(f='ChargeResolutionVariationViewer.input_path',
                        O='ChargeResolutionVariationPlotter.output_path',
                        ))
    classes = List([ChargeResolutionCalculator,
                    ChargeResolutionVariationPlotter
                    ])

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.calculator = None
        self.plotter = None

    def setup(self):
        self.log_format = "%(levelname)s: %(message)s [%(name)s.%(funcName)s]"
        kwargs = dict(config=self.config, tool=self)

        self.calculator = ChargeResolutionCalculator(**kwargs)
        self.plotter = ChargeResolutionVariationPlotter(**kwargs)

    def start(self):
        self.calculator.load(self.input_path)

    def finish(self):
        hist = self.calculator.variation_hist
        xedges = self.calculator.variation_xedges
        yedges = self.calculator.variation_yedges
        self.plotter.plot_hist(hist, xedges, yedges)


def main():
    exe = ChargeResolutionVariationViewer()
    exe.run()

if __name__ == '__main__':
    main()



from ctapipe.core import Component
from time import sleep

class Inverse(Component):
    """Add class represents a Stage for pipeline.
    It returns inverted value of received value
    """
    def init(self):
        self.log.debug("--- Add init ---")
        return True

    def run(self, input):
        sleep(.5)
        if input:
            self.log.debug("Inverse receive {}".format(input))
            return int(input) * -1

    def finish(self):
        self.log.debug("--- Add finish ---")
        pass



from copy import deepcopy

from ctapipe.core.component import Component
from abc import abstractmethod


class Factory(Component):
    # TODO: obtain traits automatically for the classes that can be produced
    """
    A base class for all class factories that exist in the `Tools`/`Components`
    frameworks.

    Either maunally set "subclasses" as a list of the classes that you wish
    to be inside the factory, or set it to
    Factory.child_subclasses(ParentClass). Using that static method will
    obtain all the lowest-level child classes from the parent class.

    You must manually specify the discriminator trait (which is what you
    set at run-time to choose the product you wish to obtain from the factory)
    and the traits of all the classes that are possible to obtain from the
    factory. Perhaps in the future these can be obtained automatically.

    You must also return the factory name in get_factory_name(), and the
    discriminator in get_product_name() in your custom Factory class.

    To then obtain the product class from the factory, use 'get_class()",
    which can be used to then initialise the class. The correct traits
    from the correspoding product are set automatically from the factory.

    .. code:: python

        from ctapipe.core import Factory
        from traitlets import (Integer, Float, List, Dict, Unicode)

        class MyFactory(Factory):
            name = "myfactory"
            description = "do some things and stuff"

            subclasses = Factory.child_subclasses(ParentClass)
            subclass_names = [c.__name__ for c in subclasses]

            discriminator = Unicode('DefaultProduct',
                             help='Product to obtain: {}'
                             .format(subclass_names)).tag(config=True)

            # Product classes traits
            # Would be nice to have these automatically set...!
            product_trait1 = Int(7, help="").tag(config=True)
            product_trait2 = Int(7, help="").tag(config=True)

            def get_factory_name(self):
                return self.name

            def get_product_name(self):
                return self.discriminator
    """
    subclasses = None  # Set to all_subclasses(ParentClass) in factory
    subclass_names = None  # Set to [c.__name__ for c in subclasses] in factory

    def __init__(self, config, tool, **kwargs):
        """
        Base Factory class

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        kwargs

        """
        super().__init__(config=config, parent=tool, **kwargs)
        self.product = None

    @staticmethod
    def child_subclasses(cls):
        """
        Return all base subclasses of a parent class. Finds the bottom level
        subclasses that have no further children.

        Parameters
        ----------
        cls : class
            high level class object that contains the desired children classes

        Returns
        -------
        list
            list of bottom level subclasses

        """
        family = cls.__subclasses__() + [g for s in cls.__subclasses__()
                                         for g in Factory.child_subclasses(s)]
        children = [g for g in family if not g.__subclasses__()]
        return children

    @abstractmethod
    def get_product_name(self):
        """
        Abstract method to be implemented in child factory.
        Simply return the discriminator traitlet.
        """

    @abstractmethod
    def get_factory_name(self):
        """
        Abstract method to be implemented in child factory.
        Simply return the name of the factory.
        """

    def get_class(self):
        """
        Obtain the class constructor for the specified product name.

        Returns
        -------
        product : class

        """
        subclass_dict = dict(zip(self.subclass_names, self.subclasses))
        self.log.info("Obtaining {} from {}".format(self.get_product_name(),
                                                    self.get_factory_name()))
        try:
            product = subclass_dict[self.get_product_name()]
        except KeyError:
            self.log.exception('No product found with name "{}" for '
                               'factory.'.format(self.get_product_name()))
            raise

        # Copy factory traits to product
        c = self.__dict__['_trait_values']['config']
        c[product.name] = deepcopy(c[self.get_factory_name()])
        items = self.__dict__['_trait_values'].items()
        for key, values in items:
            if key != 'config' and key != 'parent':
                c[product.name][key] = values
        keys = list(c[product.name].keys())
        for key in keys:
            if key not in product.class_trait_names():
                del c[product.name][key]
        return product



from traitlets import (Int, Integer, Float, Unicode, Enum, Long, List,
                       Bool, CRegExp, Dict, TraitError, observe,
                       CaselessStrEnum)
from traitlets.config import boolean_flag as flag

__all__ = ['Int', 'Integer', 'Float', 'Unicode', 'Enum', 'Long', 'List',
           'Bool', 'CRegExp', 'Dict', 'flag', 'TraitError', 'observe',
           'CaselessStrEnum']



from tempfile import TemporaryDirectory
from ctapipe.reco.event_classifier import *
from sklearn.neural_network import MLPClassifier
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
import pytest


def test_pipeline_classifier():
    cam_id_list = ["FlashCam", "ASTRICam"]
    feature_list = {"FlashCam": [[1, 10], [2, 20], [3, 30], [0.9, 9],
                                 [10, 1], [20, 2], [30, 3], [9, 0.9]],
                    "ASTRICam": [[10, 1], [20, 2], [30, 3], [9, 0.9],
                                 [1, 10], [2, 20], [3, 30], [0.9, 9]]}
    target_list = {"FlashCam": ["a", "a", "a", "a", "b", "b", "b", "b"],
                   "ASTRICam": ["a", "a", "a", "a", "b", "b", "b", "b"]}

    estimators = [('scaler', StandardScaler()),
                  ('clf', MLPClassifier(max_iter=400))]

    clf = EventClassifier(classifier=Pipeline, steps=estimators,
                          cam_id_list=cam_id_list)
    clf.fit(feature_list, target_list)

    prediction = clf.predict_by_event([{"ASTRICam": [[10, 1]]},
                                       {"ASTRICam": [[2, 20]]},
                                       {"ASTRICam": [[3, 30]]}])
    assert (prediction == ["a", "b", "b"]).all()

    prediction = clf.predict_by_event([{"FlashCam": [[10, 1]]},
                                       {"FlashCam": [[2, 20]]},
                                       {"FlashCam": [[3, 30]]}])
    assert (prediction == ["b", "a", "a"]).all()


def test_prepare_model_MLP():
    cam_id_list = ["FlashCam", "ASTRICam"]
    feature_list = {"FlashCam": [[1, 10], [2, 20], [3, 30], [0.9, 9],
                                 [10, 1], [20, 2], [30, 3], [9, 0.9]],
                    "ASTRICam": [[10, 1], [20, 2], [30, 3], [9, 0.9],
                                 [1, 10], [2, 20], [3, 30], [0.9, 9]]}
    target_list = {"FlashCam": ["a", "a", "a", "a", "b", "b", "b", "b"],
                   "ASTRICam": ["a", "a", "a", "a", "b", "b", "b", "b"]}

    clf = EventClassifier(classifier=MLPClassifier, cam_id_list=cam_id_list, max_iter=400)
    scaled_features, scaler = EventClassifier.scale_features(cam_id_list, feature_list)

    # clf.fit(feature_list, target_list)
    clf.fit(scaled_features, target_list)
    return clf, cam_id_list, scaler

def test_fit_save_load_MLP():
    clf, cam_id_list, scaler = test_prepare_model_MLP()
    with TemporaryDirectory() as d:
        temp_path = "/".join([d, "reg_{cam_id}.pkl"])
        clf.save(temp_path)
        clf = EventClassifier.load(temp_path, cam_id_list)
        return clf, cam_id_list, scaler

def test_predict_by_event_MLP():
    clf, cam_id_list, scaler = test_fit_save_load_MLP()

    x = scaler['ASTRICam'].transform(np.array([10, 1], dtype=float).reshape(1, -1))
    x = {'ASTRICam': x}
    y = scaler['ASTRICam'].transform(np.array([2, 20], dtype=float).reshape(1, -1))
    y = {'ASTRICam': y}
    z = scaler['ASTRICam'].transform(np.array([3, 30], dtype=float).reshape(1, -1))
    z = {'ASTRICam': z}
    prediction = clf.predict_by_event([x, y, z])
    assert (prediction == ["a", "b", "b"]).all()

    x = scaler['FlashCam'].transform(np.array([10, 1], dtype=float).reshape(1, -1))
    x = {'FlashCam': x}
    y = scaler['FlashCam'].transform(np.array([2, 20], dtype=float).reshape(1, -1))
    y = {'FlashCam': y}
    z = scaler['FlashCam'].transform(np.array([3, 30], dtype=float).reshape(1, -1))
    z = {'FlashCam': z}
    prediction = clf.predict_by_event([x, y, z])
    assert (prediction == ["b", "a", "a"]).all()

def test_prepare_model():
    cam_id_list = ["FlashCam", "ASTRICam"]
    feature_list = {"FlashCam": [[1, 10], [2, 20], [3, 30], [0.9, 9],
                                 [10, 1], [20, 2], [30, 3], [9, 0.9]],
                    "ASTRICam": [[10, 1], [20, 2], [30, 3], [9, 0.9],
                                 [1, 10], [2, 20], [3, 30], [0.9, 9]]}
    target_list = {"FlashCam": ["a", "a", "a", "a", "b", "b", "b", "b"],
                   "ASTRICam": ["a", "a", "a", "a", "b", "b", "b", "b"]}

    clf = EventClassifier(cam_id_list=cam_id_list)
    clf.fit(feature_list, target_list)
    return clf, cam_id_list

def test_fit_save_load():
    clf, cam_id_list = test_prepare_model()
    with TemporaryDirectory() as d:
        temp_path = "/".join([d, "reg_{cam_id}.pkl"])
        clf.save(temp_path)
        clf = EventClassifier.load(temp_path, cam_id_list)
        return clf, cam_id_list


def test_predict_by_event():
    clf, cam_id_list = test_fit_save_load()
    prediction = clf.predict_by_event([{"ASTRICam": [[10, 1]]},
                                       {"ASTRICam": [[2, 20]]},
                                       {"ASTRICam": [[3, 30]]}])
    assert (prediction == ["a", "b", "b"]).all()

    prediction = clf.predict_by_event([{"FlashCam": [[10, 1]]},
                                       {"FlashCam": [[2, 20]]},
                                       {"FlashCam": [[3, 30]]}])
    assert (prediction == ["b", "a", "a"]).all()



import numpy as np
from scipy.optimize import minimize
import scipy.constants as const
from scipy.stats import norm
from astropy.units import Quantity

__all__ = [
    'kundu_chaudhuri_circle_fit',
    'psf_likelihood_fit',
    'impact_parameter_chisq_fit',
    'mirror_integration_distance',
    'expected_pixel_light_content',
    'radial_light_intensity',
    'efficiency_fit',
]


def cherenkov_integral(lambda1, lambda2):
    ''' integral of int_lambda1^lambda2 lambda^-2 dlambda '''
    return 1 / lambda1 - 1 / lambda2


def kundu_chaudhuri_circle_fit(x, y, weights):
    '''
    Fast, analytic calculation of circle center and radius for 
    weighted data using method given in [chaudhuri93]_

    Parameters
    ----------
    x: array-like or astropy quantity
        x coordinates of the points
    y: array-like or astropy quantity
        y coordinates of the points
    weights: array-like
        weights of the points

    '''

    weights_sum = np.sum(weights)
    mean_x = np.sum(x * weights) / weights_sum
    mean_y = np.sum(y * weights) / weights_sum

    a1 = np.sum(weights * (x - mean_x) * x)
    a2 = np.sum(weights * (y - mean_y) * x)

    b1 = np.sum(weights * (x - mean_x) * y)
    b2 = np.sum(weights * (y - mean_y) * y)

    c1 = 0.5 * np.sum(weights * (x - mean_x) * (x**2 + y**2))
    c2 = 0.5 * np.sum(weights * (y - mean_y) * (x**2 + y**2))

    center_x = (b2 * c1 - b1 * c2) / (a1 * b2 - a2 * b1)
    center_y = (a2 * c1 - a1 * c2) / (a2 * b1 - a1 * b2)

    radius = np.sqrt(np.sum(
        weights * ((center_x - x)**2 + (center_y - y)**2),
    ) / weights_sum)

    return radius, center_x, center_y


def _psf_neg_log_likelihood(params, x, y, weights):
    '''
    Negative log-likelihood for a gaussian ring profile

    Parameters
    ----------
    params: 4-tuple
        the fit parameters: (radius, center_x, center_y, std)
    x: array-like
        x coordinates
    y: array-like
        y coordinates
    weights: array-like
        weights for the (x, y) points

    This will usually be x and y coordinates and pe charges of camera pixels
    '''
    radius, center_x, center_y, sigma = params
    pixel_distance = np.sqrt((center_x - x)**2 + (center_y - y)**2)

    return np.sum(
        (np.log(sigma) + 0.5 * ((pixel_distance - radius) / sigma)**2) * weights
    )


def psf_likelihood_fit(x, y, weights):
    '''
    Do a likelihood fit using a ring with gaussian profile.
    Uses the kundu_chaudhuri_circle_fit for the initial guess



    Parameters
    ----------
    x: array-like or astropy quantity
        x coordinates of the points
    y: array-like or astropy quantity
        y coordinates of the points
    weights: array-like
        weights of the points

    This will usually be x and y coordinates and pe charges of camera pixels

    Returns
    -------
    radius: astropy-quantity
        radius of the ring
    center_x: astropy-quantity
        x coordinate of the ring center
    center_y: astropy-quantity
        y coordinate of the ring center
    std: astropy-quantity
        standard deviation of the gaussian profile (indictor for the ring width)
    '''

    x = Quantity(x).decompose()
    y = Quantity(y).decompose()
    assert x.unit == y.unit
    unit = x.unit
    x = x.value
    y = y.value

    start_r, start_x, start_y = kundu_chaudhuri_circle_fit(x, y, weights)

    result = minimize(
        _psf_neg_log_likelihood,
        x0=(start_r, start_x, start_y, 5e-3),
        args=(x, y, weights),
        method='L-BFGS-B',
        bounds=[
            (0, None),      # radius should be positive
            (None, None),
            (None, None),
            (0, None),      # std should be positive
        ],
    )

    if not result.success:
        result.x = np.full_like(result.x, np.nan)

    return result.x * unit


def impact_parameter_chisq_fit(
        pixel_x,
        pixel_y,
        weights,
        radius,
        center_x,
        center_y,
        mirror_radius,
        bins=30,
        ):
    '''
    Impact parameter calculation for a ring fit before.
    This is fitting the theoretical angular light distribution to the
    observed binned angular light distribution using least squares

    Parameters
    ----------
    pixel_x: array-like
        x coordinates of the pixels
    pixel_y: array-like
        y coordinates of the pixels
    weights: array-like
        the weights for the pixel, usually this should be the pe_charge
    radius: float
        ring radius, for example estimated by psf_likelihood_fit
    center_x: float
        x coordinate of the ring center, for example estimated by psf_likelihood_fit
    center_y: float
        y coordinate of the ring center, for example estimated by psf_likelihood_fit
    mirror_radius: float
        the radius of the telescope mirror (circle approximation)
    bins: int
        how many bins to use for the binned fit
    '''

    phi = np.arctan2(pixel_y - center_y, pixel_x - center_x)
    hist, edges = np.histogram(phi, bins=bins, range=[-np.pi, np.pi], weights=weights)
    bin_centers = 0.5 * (edges[:-1] + edges[1:])

    result = minimize(
        _impact_parameter_chisq,
        x0=(mirror_radius / 2, bin_centers[np.argmax(hist)], 1),
        args=(bin_centers, hist, mirror_radius),
        method='L-BFGS-B',
        bounds=[
            (0, None),         # impact parameter should be positive
            (-np.pi, np.pi),   # orientation angle should be in -pi to pi
            (0, None),         # scale should be positive
        ],
    )

    if not result.success:
        result.x = np.full_like(result.x, np.nan)

    imp_par, phi_max, scale = result.x

    return imp_par, phi_max


def mirror_integration_distance(phi, phi_max, impact_parameter, mirror_radius):
    '''
    Calculate the distance the muon light went across the telescope mirror
    Function (6) from G. Vacanti et. al., Astroparticle Physics 2, 1994, 1-11

    Parameters
    ----------
    phi: float or array-like
        the orientation angle on the ring
    phi_max: float
        position of the light maximum
    impact_parameter: float
        distance of the muon impact point from the mirror center
    mirror_radius: float
        radius of the telescope mirror (circle approximation)

    Returns
    -------
    distance: float or array-like
    '''
    phi = phi - phi_max
    ratio = impact_parameter / mirror_radius
    radicant = 1 - ratio**2 * np.sin(phi)**2

    if impact_parameter > mirror_radius:
        distance = np.empty_like(phi)
        mask = np.logical_and(
            phi < np.arcsin(1 / ratio),
            phi > -np.arcsin(1 / ratio)
        )
        distance[np.logical_not(mask)] = 0
        distance[mask] = 2 * mirror_radius * np.sqrt(radicant[mask])
    else:
        distance = 2 * mirror_radius * (np.sqrt(radicant) + ratio * np.cos(phi))

    return distance


def radial_light_intensity(
        phi,
        phi_max,
        cherenkov_angle,
        impact_parameter,
        pixel_fov,
        mirror_radius,
        lambda1=300e-9,
        lambda2=900e-9,
        ):
    '''
    Amount of photons per azimuthal angle phi on the muon ring as given in
    formula (5) of [vacanti94]_ 

    Parameters
    ----------
    phi: float or array-like
        the orientation angle on the ring
    phi_max: float
        position of the light maximum
    impact_parameter: float
        distance of the muon impact point from the mirror center
    pixel_fov: float
        field of view of the camera pixels in radian
    mirror_radius: float
        radius of the telescope mirror (circle approximation)
    lambda1: float
        lower integration limit over the cherenkov spectrum in meters
    lambda2: float
        upper integration limit over the cherenkov spectrum in meters

    Returns
    -------
    light_density: float or array-like
        
    '''

    return (
        0.5 * const.fine_structure *
        cherenkov_integral(lambda1, lambda2) *
        pixel_fov / cherenkov_angle *
        np.sin(2 * cherenkov_angle) *
        mirror_integration_distance(phi, phi_max, impact_parameter, mirror_radius)
    )


def expected_pixel_light_content(
        pixel_x,
        pixel_y,
        center_x,
        center_y,
        phi_max,
        cherenkov_angle,
        impact_parameter,
        sigma_psf,
        pixel_fov,
        pixel_diameter,
        mirror_radius,
        focal_length,
        lambda1=300e-9,
        lambda2=900e-9,
        ):
    '''
    Calculate the expected light content of each pixel for a muon ring with the
    given properties

    Parameters
    ----------
    pixel_x: array-like
        x coordinates of the pixels
    pixel_y: array-like
        y coordinates of the pixels
    center_x: float
        x coordinate of the ring center, for example estimated by psf_likelihood_fit
    center_y: float
        y coordinate of the ring center, for example estimated by psf_likelihood_fit
    phi_max: float
        position of the light maximum
    cherenkov_angle: float
        cherenkov_angle of the muon light
    impact_parameter: float
        distance of the muon impact point from the mirror center
    sigma_pdf: float
        standard deviation for the gaussian profile of the ring
    pixel_fov: float
        field of view of the camera pixels in radian
    pixel_diameter: float
        diameter of the pixels
    mirror_radius: float
        radius of the telescope mirror (circle approximation)
    focal_length: float
        focal legth of the telescope
    lambda1: float
        lower integration limit over the cherenkov spectrum in meters
    lambda2: float
        upper integration limit over the cherenkov spectrum in meters

    Returns
    -------
    pe_charge: array-like
        number of photons for each pixel given in pixel_x, pixel_y
    '''
    phi = np.arctan2(pixel_y - center_y, pixel_x - center_x)
    pixel_r = np.sqrt((pixel_x - center_x)**2 + (pixel_y - center_y)**2)
    ring_radius = cherenkov_angle * focal_length

    light = radial_light_intensity(
        phi, phi_max,
        cherenkov_angle, impact_parameter,
        pixel_fov, mirror_radius, lambda1, lambda2
    )

    result = light * pixel_diameter * norm.pdf(pixel_r, ring_radius, sigma_psf)
    return result


def efficiency_fit(
        pe_charge,
        pixel_x,
        pixel_y,
        pixel_fov,
        pixel_diameter,
        mirror_radius,
        focal_length,
        lambda1=300e-9,
        lambda2=900e-9,
        ):
    '''
    Estimate optical efficiency for a muon ring using method of [mitchell15]_.
    This is performing several steps:
    
    1. fit r, x, y and width with the psf_likelihood_fit
    2. fit impact parameter with impact_parameter_chisq_fit
    3. calculate the theoretically expected light contents for each pixel 
       for the estimated parameters
    4. calculate the ratio between the observed and the expected number
       of photons.
        
    Parameters
    ----------
    pe_charge: array-like
        pe_charges of the pixels
    pixel_x: array-like
        x coordinates of the pixels
    pixel_y: array-like
        y coordinates of the pixels
    pixel_fov: float
        field of view of the camera pixels in radian
    pixel_diameter: float
        diameter of the pixels
    mirror_radius: float
        radius of the telescope mirror (circle approximation)
    focal_length: float
        focal legth of the telescope
    lambda1: float
        lower integration limit over the cherenkov spectrum in meters
    lambda2: float
        upper integration limit over the cherenkov spectrum in meters

    Returns
    -------
    radius
    center_x
    center_y
    sigma_psf
    imp_par
    phi_max
    efficiency
    '''

    radius, center_x, center_y, sigma_psf = psf_likelihood_fit(
        pixel_x, pixel_y, pe_charge
    )

    imp_par, phi_max = impact_parameter_chisq_fit(
        pixel_x=pixel_x,
        pixel_y=pixel_y,
        weights=pe_charge,
        center_x=center_x,
        center_y=center_y,
        radius=radius,
        mirror_radius=mirror_radius,
        bins=30,
    )

    expected_light = expected_pixel_light_content(
        pixel_x=pixel_x,
        pixel_y=pixel_y,
        center_x=center_x,
        center_y=center_y,
        phi_max=phi_max,
        cherenkov_angle=radius / focal_length,
        impact_parameter=imp_par,
        sigma_psf=sigma_psf,
        pixel_fov=pixel_fov,
        pixel_diameter=pixel_diameter,
        mirror_radius=mirror_radius,
        focal_length=focal_length,
        lambda1=lambda1,
        lambda2=lambda2,
    )

    efficiency = np.sum(pe_charge) / np.sum(expected_light)

    return radius, center_x, center_y, sigma_psf, imp_par, phi_max, efficiency


def _impact_parameter_chisq(params,  phi, hist, mirror_radius):
    ''' function (6) from G. Vacanti et. al., Astroparticle Physics 2, 1994, 1-11 '''

    imp_par, phi_max, scale = params
    theory = mirror_integration_distance(phi, phi_max, imp_par, mirror_radius)

    return np.sum((hist - scale * theory)**2)



"""Module to handle plotting of camera event specific items, e.g.
camera images and waveforms.
"""

from matplotlib import pyplot as plt

from ctapipe.visualization import CameraDisplay

from astropy import units as u
from astropy import log


class CameraPlotter:
    """
    Plotter object for basic items that are camera event specific.

    WARNING: a new CameraPlotter object must be created for each event

    Attributes
    ----------
    event : container
        A `ctapipe` event container
    """

    def __init__(self, event):
        """
        Parameters
        ----------
        event : container
            A `ctapipe` event container
        """
        self.event = event
        self.cameradisplay_dict = {}

    def get_geometry(self, tel):
        return self.event.inst.subarray.tel[tel].camera

    def draw_camera(self, tel, data, axes=None):
        """
        Draw a camera image using the correct geometry.

        Parameters
        ----------
        tel : int
            The telescope you want drawn.
        data : `np.array`
            1D array with length equal to npix.
        axes : `matplotlib.axes.Axes`
            A matplotlib axes object to plot on, or None to create a new one.

        Returns
        -------
        `ctapipe.visualization.CameraDisplay`
        """

        geom = self.get_geometry(tel)
        axes = axes if axes is not None else plt.gca()
        camera = CameraDisplay(geom, ax=axes)
        camera.image = data
        camera.cmap = plt.cm.viridis
        # camera.add_colorbar(ax=axes, label="Amplitude (ADC)")
        # camera.set_limits_percent(95)  # autoscale
        return camera

    def draw_camera_pixel_ids(self, tel, pixels, axes=None):
        """
        Draw the pixel_ids on top of a camera image

        Parameters
        ----------
        tel : int
            The telescope you want drawn.
        pixels : list
            A list of the pixel IDs you want drawing
        axes : `matplotlib.axes.Axes`
            A matplotlib axes object to plot on, or None to create a new one.
        """

        geom = self.get_geometry(tel)
        axes = axes if axes is not None else plt.gca()
        log.info("Annotating with pixel_ids")
        for pix in pixels:
            x = u.Quantity(geom.pix_x).value[pix]
            y = u.Quantity(geom.pix_y).value[pix]
            axes.text(x, y, pix, fontsize=2, ha='center')

    def draw_camera_pixel_annotation(self, tel, p0, p1, axes=None):
        """
        Draw annotations for two pixels, used for camera images that are
        accompanied by two waveforms

        Parameters
        ----------
        tel : int
            The telescope you want drawn.
        p0 : int
            First pixel_id
        p1 : int
            Second pixel_id
        axes : `matplotlib.axes.Axes`
            A matplotlib axes object to plot on, or None to create a new one.
        """

        geom = self.get_geometry(tel)
        axes = axes if axes is not None else plt.gca()
        axes.annotate("Pixel: {}".format(p0),
                      xy=(u.Quantity(geom.pix_x).value[p0],
                          u.Quantity(geom.pix_y).value[p0]),
                      xycoords='data', xytext=(0.05, 0.98),
                      textcoords='axes fraction',
                      arrowprops=dict(facecolor='red', width=2, alpha=0.4),
                      horizontalalignment='left', verticalalignment='top')
        axes.annotate("Pixel: {}".format(p1),
                      xy=(u.Quantity(geom.pix_x).value[p1],
                          u.Quantity(geom.pix_y).value[p1]),
                      xycoords='data', xytext=(0.05, 0.94),
                      textcoords='axes fraction',
                      arrowprops=dict(facecolor='orange', width=2, alpha=0.4),
                      horizontalalignment='left', verticalalignment='top')

    @staticmethod
    def draw_waveform(data, axes=None):
        """
        Draw the waveform from a pixel.

        Parameters
        ----------
        data : `np.array`
            1D array with length equal to a waveforms sample size.
        axes : `matplotlib.axes.Axes`
            A matplotlib axes object to plot on, or None to create a new one.

        Returns
        -------
        `matplotlib.lines.Line2D`

        """

        axes = axes if axes is not None else plt.gca()
        waveform, = axes.plot(data)
        axes.set_xlabel("Time (ns)")
        axes.set_ylabel("Amplitude (ADC)")
        return waveform

    @staticmethod
    def draw_waveform_positionline(t, axes=None):
        """
        Draw a vertical line on a waveform. Intended to be used for showing
        postion along a waveform for a animation or the event.
        Parameters
        ----------
        t : float
            X position (time) for the line to be drawn at on the waveform.
        axes : `matplotlib.axes.Axes`
            A matplotlib axes object to plot on, or None to create a new one.

        Returns
        -------
        `matplotlib.lines.Line2D`

        """
        axes = axes if axes is not None else plt.gca()
        line, = axes.plot([t, t], axes.get_ylim(), color='r', alpha=1)
        return line
